/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id$
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.role.util;

import it.tidalwave.netbeans.role.util.impl.LookupAllProvider;
import it.tidalwave.netbeans.role.util.impl.LookupProvider;
import java.lang.reflect.Type;
import javax.annotation.Nonnull;
import javax.annotation.Resource;
import javax.inject.Inject;
import java.lang.reflect.Field;
import java.util.Comparator;
import java.util.Set;
import java.util.TreeSet;
import org.openide.util.Lookup;
import org.openide.util.lookup.ProxyLookup;
import it.tidalwave.util.logging.Logger;
import java.lang.reflect.ParameterizedType;
import java.util.Collection;
import javax.inject.Provider;

/***********************************************************************************************************************
 *
 * This class represents the abstract capability of loading data somewhere.
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public final class RoleInjector
  {
    private final static String CLASS = RoleInjector.class.getName();
    private final static Logger logger = Logger.getLogger(CLASS);

    private RoleInjector()
      {
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public static void injectLookup (final @Nonnull Object role,
                                     final @Nonnull Lookup injectedLookup)
      {
        injectLookup(role, role.getClass(), new ProxyLookup(injectedLookup, Lookup.getDefault()));
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public static Set<Class<?>> findDependencies (final @Nonnull Object role)
      {
        return findDependencies(role, role.getClass());
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @edu.umd.cs.findbugs.annotations.SuppressWarnings("REC_CATCH_EXCEPTION")
    private static void injectLookup (final @Nonnull Object role,
                                      final @Nonnull Class<?> clazz,
                                      final @Nonnull Lookup injectedLookup)
      {
        logger.finest("injectLookup(%s, %s, ...)", role, clazz);
        assert injectedLookup != null : "injectedLookup is null";

        final Class<?> superclass = clazz.getSuperclass();

        if (superclass != null)
          {
            injectLookup(role, superclass, injectedLookup);
          }

        for (final Field field : clazz.getDeclaredFields())
          {
            if ((field.getAnnotation(Resource.class) != null) || (field.getAnnotation(Inject.class) != null))
              {
                logger.finest(">>>> injecting field %s...", field);

                Class<?> fieldType = field.getType();

                try
                  {
                    Object toInject = null;

                    if (fieldType.equals(Lookup.class))
                      {
                        toInject = injectedLookup;
                      }
                    else if (fieldType.equals(Provider.class))
                      {
                        final ParameterizedType genericType = (ParameterizedType)field.getGenericType();
                        final Type x = genericType.getActualTypeArguments()[0];
                        toInject = new LookupProvider(injectedLookup, (Class<?>)x);
                        logger.info(">>>>>>>> Injecting provider for %s", x);
                      }
                    else
                      {
                        toInject = createProvider(fieldType, field, injectedLookup).get();
                      }

                    if (toInject == null)
                      {
                        logger.info("Can't lookup resource: %s", fieldType);
                      }

                    logger.info(">>>>>>>> injected object is: %s", toInject);
                    field.setAccessible(true);
                    field.set(role, toInject);
                  }
                catch (Exception e)
                  {
                    logger.severe("While injecting %s to %s: %s", fieldType, role, e);
                    logger.throwing(CLASS, "injectLookup()", e);
                  }
              }
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static Provider<?> createProvider (final @Nonnull Class<?> fieldType,
                                               final @Nonnull Field field,
                                               final @Nonnull Lookup lookup)
      {
        if (!Collection.class.isAssignableFrom(fieldType))
          {
            return new LookupProvider(lookup, fieldType);
          }
        else
          {
            final ParameterizedType genericType = (ParameterizedType)field.getGenericType();
            final Type itemClass = genericType.getActualTypeArguments()[0];
            return new LookupAllProvider(lookup, (Class<?>)itemClass);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static Set<Class<?>> findDependencies (final @Nonnull Object role,
                                                   final @Nonnull Class<?> clazz)
      {
        final Set<Class<?>> result = new TreeSet<Class<?>>(new Comparator<Class<?>>()
          {
            @Override
            public int compare (final @Nonnull Class<?> c1, final @Nonnull Class<?> c2)
              {
                return c1.getName().compareTo(c2.getName());
              }
          });
        
        final Class<?> superclass = clazz.getSuperclass();

        if (superclass != null)
          {
            result.addAll(findDependencies(role, superclass));
          }

        for (final Field field : clazz.getDeclaredFields())
          {
            if (field.getAnnotation(Resource.class) != null)
              {
                final Class<?> fieldType = field.getType();
                result.add(fieldType);
              }
          }

        return result;
      }
  }
